@@ -44,13 +44,6 @@ function check_nonroot
 
 	$MAKE || return
 
-	sudo modprobe dm-crypt
-	sudo modprobe dm-verity
-	sudo modprobe dm-integrity
-	uname -a
-	sudo dmsetup version
-	sudo dmsetup targets
-
 	make check
 }
 
@@ -69,13 +62,6 @@ function check_root
 
 	$MAKE || return
 
-	sudo modprobe dm-crypt
-	sudo modprobe dm-verity
-	sudo modprobe dm-integrity
-	uname -a
-	sudo dmsetup version
-	sudo dmsetup targets
-
 	# FIXME: we should use -E option here
 	sudo make check
 }
