@@ -153,7 +153,7 @@ int crypt_wipe_device(struct crypt_device *cd,
 	/* FIXME: if wipe_block_size < bsize, then a wipe is highly ineffective */
 
 	/* Everything must be aligned to SECTOR_SIZE */
-	if ((offset % SECTOR_SIZE) || (length % SECTOR_SIZE) || (wipe_block_size % SECTOR_SIZE))
+	if (MISALIGNED_512(offset) || MISALIGNED_512(length) || MISALIGNED_512(wipe_block_size))
 		return -EINVAL;
 
 	devfd = device_open(device, O_RDWR);
@@ -161,9 +161,12 @@ int crypt_wipe_device(struct crypt_device *cd,
 		return errno ? -errno : -EINVAL;
 
 	r = device_size(device, &dev_size);
-	if (r)
+	if (r || dev_size == 0)
 		goto out;
 
+	if (dev_size < length)
+		length = 0;
+
 	if (length) {
 		if ((dev_size <= offset) || (dev_size - offset) < length) {
 			r = -EINVAL;
@@ -177,7 +180,7 @@ int crypt_wipe_device(struct crypt_device *cd,
 		goto out;
 
 	if (lseek64(devfd, offset, SEEK_SET) < 0) {
-		log_err(cd, "Cannot seek to device offset.\n");
+		log_err(cd, "Cannot seek to device offset.");
 		r = -EINVAL;
 		goto out;
 	}
@@ -201,7 +204,7 @@ int crypt_wipe_device(struct crypt_device *cd,
 		r = wipe_block(devfd, pattern, sf, bsize, alignment,
 			       wipe_block_size, offset, &need_block_init);
 		if (r) {
-			log_err(cd, "Device wipe error, offset %" PRIu64 ".\n", offset);
+			log_err(cd, "Device wipe error, offset %" PRIu64 ".", offset);
 			break;
 		}
 
@@ -213,7 +216,7 @@ int crypt_wipe_device(struct crypt_device *cd,
 		}
 	}
 
-	fsync(devfd);
+	device_sync(device, devfd);
 out:
 	close(devfd);
 	free(sf);
